'''
    Step Fetch
'''
import json
import time
import httpx
from httpx import Response

from api_pipe.api import ApiPipe
from api_pipe.error import FetchMaxAttempts

def run_step(self: ApiPipe) -> object:
    '''
        Run step

        object is an ApiPipe object
        (Due to circular imports )
    '''
    self.log.info(
        f"Fetching from {self.url}"
    )

    for attempt in range(self.retries_max_retries):
        self.log.debug(
            f"Attempt: {attempt + 1} of {self.retries_max_retries}"
        )
        try:
            if self.client:
                response: Response = self.client.get(
                    str(self.url),
                    timeout=self.timeout,
                    headers=self.headers
                )
            else:
                response = httpx.get(
                    str(self.url),
                    timeout=self.timeout,
                    headers=self.headers
                )

            response.raise_for_status()
            self.data = response.text
            break
        except Exception as e:
            if attempt == self.retries_max_retries - 1:
                self.log.error(
                    f"{self.log_unique_name}. Attempted max times {self.retries_max_retries}"
                )
                raise FetchMaxAttempts(
                    self.url,
                    self.retries_max_retries
                ) from e
            sleeping_time = self.retries_initial_delay * (self.retries_backoff_factor ** attempt)
            self.log.warning(
                f"{self.log_unique_name}. "
                f"\nRetrying fetch for URL {self.url}. Attempt: {attempt + 1} of {self.retries_max_retries}."
                f"\nError: {e}"
                f"\nSleeping for {sleeping_time} seconds"
            )

            for handler in self.log.handlers:
                handler.flush()

            time.sleep(
                sleeping_time
            )

    self.type = "text"

    self._log_step_to_file(
        "fetch",
        f"GET {self.url}\n{json.dumps(self.headers)}",
        self.data
    )

    return self
